//-----------------------------------------------------------------------------
// ExternalLobVar.c
//   Defines the routines for handling LOB variables external to this module.
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// external LOB type
//-----------------------------------------------------------------------------
typedef struct {
    PyObject_HEAD
    udt_LobVar *lobVar;
    unsigned pos;
    unsigned internalFetchNum;
} udt_ExternalLobVar;


//-----------------------------------------------------------------------------
// Declaration of external LOB variable functions.
//-----------------------------------------------------------------------------
static void ExternalLobVar_Free(udt_ExternalLobVar*);
static PyObject *ExternalLobVar_Str(udt_ExternalLobVar*);
static PyObject *ExternalLobVar_GetAttr(udt_ExternalLobVar*, PyObject*);
static PyObject *ExternalLobVar_Size(udt_ExternalLobVar*, PyObject*);
static PyObject *ExternalLobVar_Read(udt_ExternalLobVar*, PyObject*);
static PyObject *ExternalLobVar_Write(udt_ExternalLobVar*, PyObject*);
static PyObject *ExternalLobVar_Trim(udt_ExternalLobVar*, PyObject*);
static PyObject *ExternalLobVar_Reduce(udt_ExternalLobVar*);


//-----------------------------------------------------------------------------
// Python type declaration
//-----------------------------------------------------------------------------
static PyTypeObject g_ExternalLobVarType = {
    PyObject_HEAD_INIT(NULL)
    0,					// ob_size
    "cx_Oracle.LOB",			// tp_name
    sizeof(udt_ExternalLobVar),		// tp_basicsize
    0,					// tp_itemsize
    (destructor) ExternalLobVar_Free,	// tp_dealloc
    0,					// tp_print
    0,					// tp_getattr
    0,					// tp_setattr
    0,					// tp_compare
    0,					// tp_repr
    0,					// tp_as_number
    0,					// tp_as_sequence
    0,					// tp_as_mapping
    0,					// tp_hash
    0,					// tp_call
    (reprfunc) ExternalLobVar_Str,	// tp_str
    (getattrofunc) ExternalLobVar_GetAttr,
                                        // tp_getattro
    0,					// tp_setattro
    0,					// tp_as_buffer
    Py_TPFLAGS_DEFAULT,			// tp_flags
    0					// tp_doc
};


//-----------------------------------------------------------------------------
// declaration of methods for Python type "ExternalLOBVar"
//-----------------------------------------------------------------------------
static PyMethodDef g_ExternalLobVarMethods[] = {
    { "size", (PyCFunction) ExternalLobVar_Size, METH_NOARGS },
    { "read", (PyCFunction) ExternalLobVar_Read, METH_VARARGS },
    { "write", (PyCFunction) ExternalLobVar_Write, METH_VARARGS },
    { "trim", (PyCFunction) ExternalLobVar_Trim, METH_VARARGS },
    { "__reduce__", (PyCFunction) ExternalLobVar_Reduce, METH_NOARGS },
    { NULL, NULL }
};


//-----------------------------------------------------------------------------
// ExternalLobVar_New()
//   Create a new external LOB variable.
//-----------------------------------------------------------------------------
PyObject *ExternalLobVar_New(
    udt_LobVar *var,			// variable to encapsulate
    unsigned pos)			// position in array to encapsulate
{
    udt_ExternalLobVar *newVar;

    newVar = PyObject_NEW(udt_ExternalLobVar, &g_ExternalLobVarType);
    if (!newVar)
        return NULL;
    newVar->pos = pos;
    newVar->internalFetchNum = var->internalFetchNum;
    Py_INCREF(var);
    newVar->lobVar = var;

    return (PyObject*) newVar;
}


//-----------------------------------------------------------------------------
// ExternalLobVar_Free()
//   Free an external LOB variable.
//-----------------------------------------------------------------------------
static void ExternalLobVar_Free(
    udt_ExternalLobVar *var)		// variable to free
{
    Py_DECREF(var->lobVar);
    PyObject_DEL(var);
}


//-----------------------------------------------------------------------------
// ExternalLobVar_Verify()
//   Verify that the external LOB var is still valid.
//-----------------------------------------------------------------------------
static int ExternalLobVar_Verify(
    udt_ExternalLobVar *var)		// variable to verify
{
    if (var->internalFetchNum != var->lobVar->internalFetchNum) {
        PyErr_SetString(g_ProgrammingErrorException,
                "LOB variable no longer valid after subsequent fetch");
        return -1;
    }
    return 0;
}


//-----------------------------------------------------------------------------
// ExternalLobVar_GetAttr()
//   Retrieve an attribute on the external LOB variable object.
//-----------------------------------------------------------------------------
static PyObject *ExternalLobVar_GetAttr(
    udt_ExternalLobVar *var,		// cursor object
    PyObject *name)			// name of attribute
{
    return Py_FindMethod(g_ExternalLobVarMethods, (PyObject*) var,
            PyString_AS_STRING(name));
}


//-----------------------------------------------------------------------------
// ExternalLobVar_InternalSize()
//   Return the size of the LOB variable for internal comsumption.
//-----------------------------------------------------------------------------
static int ExternalLobVar_InternalSize(
    udt_ExternalLobVar *var)		// variable to return the size of
{
    sword status;
    ub4 length;

    status = OCILobGetLength(var->lobVar->connection->handle,
            var->lobVar->environment->errorHandle,
            var->lobVar->data[var->pos], &length);
    if (Environment_CheckForError(var->lobVar->environment, status,
            "ExternalLobVar_InternalSize()") < 0)
        return -1;

    return length;
}


//-----------------------------------------------------------------------------
// ExternalLobVar_Value()
//   Return a portion (or all) of the data in the external LOB variable.
//-----------------------------------------------------------------------------
static PyObject *ExternalLobVar_Value(
    udt_ExternalLobVar *var,		// variable to return the size of
    int offset,				// offset into LOB
    int amount)				// amount to read from LOB
{
    ub4 length, bufferSize;
    PyObject *result;
    sword status;
    char *buffer;

    // modify the arguments
    if (offset < 0)
        offset = 1;
    if (amount < 0) {
        amount = ExternalLobVar_InternalSize(var);
        if (amount < 0)
            return NULL;
        amount = amount - offset + 1;
        if (amount <= 0)
            amount = 1;
    }
    length = amount;
    bufferSize = amount * var->lobVar->environment->maxBytesPerCharacter;

    // create a string for retrieving the value
    buffer = PyMem_Malloc(bufferSize);
    if (!buffer)
        return PyErr_NoMemory();
    status = OCILobRead(var->lobVar->connection->handle,
            var->lobVar->environment->errorHandle,
            var->lobVar->data[var->pos], &length, offset, buffer,
            bufferSize, NULL, NULL, 0, 0); 
    if (Environment_CheckForError(var->lobVar->environment, status,
            "ExternalLobVar_Value()") < 0)
        return NULL;

    // return the result
    if (var->lobVar->environment->fixedWidth)
        length = length * var->lobVar->environment->maxBytesPerCharacter;
    result = PyString_FromStringAndSize(buffer, length);
    PyMem_Free(buffer);
    return result;
}


//-----------------------------------------------------------------------------
// ExternalLobVar_Size()
//   Return the size of the data in the LOB variable.
//-----------------------------------------------------------------------------
static PyObject *ExternalLobVar_Size(
    udt_ExternalLobVar *var,		// variable to return the size of
    PyObject *args)			// arguments
{
    int length;

    if (ExternalLobVar_Verify(var) < 0)
        return NULL;
    length = ExternalLobVar_InternalSize(var);
    if (length < 0)
        return NULL;
    return PyInt_FromLong(length);
}


//-----------------------------------------------------------------------------
// ExternalLobVar_Read()
//   Return a portion (or all) of the data in the external LOB variable.
//-----------------------------------------------------------------------------
static PyObject *ExternalLobVar_Read(
    udt_ExternalLobVar *var,		// variable to return the size of
    PyObject *args)			// arguments
{
    int offset, amount;

    // offset and amount are expected, both optional
    offset = amount = -1;
    if (!PyArg_ParseTuple(args, "|ii", &offset, &amount))
        return NULL;

    if (ExternalLobVar_Verify(var) < 0)
        return NULL;
    return ExternalLobVar_Value(var, offset, amount);
}


//-----------------------------------------------------------------------------
// ExternalLobVar_Str()
//   Return all of the data in the external LOB variable.
//-----------------------------------------------------------------------------
static PyObject *ExternalLobVar_Str(
    udt_ExternalLobVar *var)		// variable to return the string for
{
    if (ExternalLobVar_Verify(var) < 0)
        return NULL;
    return ExternalLobVar_Value(var, 1, -1);
}


//-----------------------------------------------------------------------------
// ExternalLobVar_Write()
//   Write a value to the LOB variable; return the number of bytes written.
//-----------------------------------------------------------------------------
static PyObject *ExternalLobVar_Write(
    udt_ExternalLobVar *var,		// variable to perform write against
    PyObject *args)			// arguments
{
    int offset, bufferLength, length;
    char *buffer;
    sword status;

    // buffer and offset are expected, offset is optional
    offset = -1;
    if (!PyArg_ParseTuple(args, "s#|i", &buffer, &bufferLength, &offset))
        return NULL;
    if (offset < 0)
        offset = 1;

    // create a string for retrieving the value
    if (ExternalLobVar_Verify(var) < 0)
        return NULL;
    length = bufferLength;
    status = OCILobWrite(var->lobVar->connection->handle,
            var->lobVar->environment->errorHandle, var->lobVar->data[var->pos],
            (unsigned int*) &length, offset, buffer, bufferLength,
            OCI_ONE_PIECE, NULL, NULL, 0, 0);
    if (Environment_CheckForError(var->lobVar->environment, status,
            "ExternalLobVar_Write()") < 0)
        return NULL;

    // return the result
    return PyInt_FromLong(length);
}


//-----------------------------------------------------------------------------
// ExternalLobVar_Trim()
//   Trim the LOB variable to the specified length.
//-----------------------------------------------------------------------------
static PyObject *ExternalLobVar_Trim(
    udt_ExternalLobVar *var,		// variable to perform write against
    PyObject *args)			// arguments
{
    sword status;
    ub4 newSize;

    // buffer and offset are expected, offset is optional
    newSize = 0;
    if (!PyArg_ParseTuple(args, "|i", &newSize))
        return NULL;

    // create a string for retrieving the value
    if (ExternalLobVar_Verify(var) < 0)
        return NULL;
    status = OCILobTrim(var->lobVar->connection->handle,
            var->lobVar->environment->errorHandle, var->lobVar->data[var->pos],
            newSize);
    if (Environment_CheckForError(var->lobVar->environment, status,
            "ExternalLobVar_Trim()") < 0)
        return NULL;

    // return the result
    Py_INCREF(Py_None);
    return Py_None;
}


//-----------------------------------------------------------------------------
// ExternalLobVar_Reduce()
//   Method provided for pickling/unpickling of LOB variables.
//-----------------------------------------------------------------------------
static PyObject *ExternalLobVar_Reduce(
    udt_ExternalLobVar *self)		// variable to dump
{
    PyObject *result, *value;

    value = ExternalLobVar_Str(self);
    if (!value)
        return NULL;
    result = Py_BuildValue("(O(O))", &PyString_Type, value);
    Py_DECREF(value);
    return result;
}

