//-----------------------------------------------------------------------------
// cx_Oracle.c
//   Shared library for use by Python.
//-----------------------------------------------------------------------------

#include <Python.h>
#include <structmember.h>
#include <time.h>
#include <oci.h>
#include <xa.h>

// define whether or not we are building Oracle 9i
#ifdef OCI_ATTR_MODULE
#define ORACLE_10G
#endif
#if OCI_FNCODE_STMTRELEASE
#define ORACLE_9I
#endif

// define whether or not we are building with the native datetime module
#if (PY_VERSION_HEX >= 0x02040000)
#include <datetime.h>
#define NATIVE_DATETIME
#endif


// define macro for setting OCI constants
#define SET_OCI_CONSTANT(x) \
    if (PyDict_SetItemString(g_ModuleDictionary, #x, \
            PyInt_FromLong(OCI_ ## x)) < 0) \
        return;
    

//-----------------------------------------------------------------------------
// Globals
//-----------------------------------------------------------------------------
#ifdef WITH_THREAD
static PyInterpreterState *g_InterpreterState;
#endif
static PyObject *g_WarningException = NULL;
static PyObject *g_ErrorException = NULL;
static PyObject *g_InterfaceErrorException = NULL;
static PyObject *g_DatabaseErrorException = NULL;
static PyObject *g_DataErrorException = NULL;
static PyObject *g_OperationalErrorException = NULL;
static PyObject *g_IntegrityErrorException = NULL;
static PyObject *g_InternalErrorException = NULL;
static PyObject *g_ProgrammingErrorException = NULL;
static PyObject *g_NotSupportedErrorException = NULL;
static PyObject *g_ModuleDictionary = NULL;
static PyTypeObject *g_DateTimeType = NULL;


//-----------------------------------------------------------------------------
// SetException()
//   Create an exception and set it in the provided dictionary.
//-----------------------------------------------------------------------------
static int SetException(
    PyObject *dictionary,		// dictionary to modify
    PyObject **exception,		// exception to create
    char *name,				// name of the exception
    PyObject *baseException)		// exception to base exception on
{
    char buffer[100];

    sprintf(buffer, "cx_Oracle.%s", name);
    *exception = PyErr_NewException(buffer, baseException, NULL);
    if (!*exception)
        return -1;
    return PyDict_SetItemString(dictionary, name, *exception);
}


#include "Environment.c"
#ifdef ORACLE_9I
#include "SessionPool.c"
#else
#include "Connection.c"
#endif


//-----------------------------------------------------------------------------
// MakeDSN()
//   Make a data source name given the host port and SID.
//-----------------------------------------------------------------------------
static PyObject* MakeDSN(
    PyObject* self,			// passthrough argument
    PyObject* args)			// arguments to function
{
    char *host, *sid;
    int port;

    // parse arguments
    if (!PyArg_ParseTuple(args, "sis", &host, &port, &sid))
        return NULL;

    // return the formatted string
    return PyString_FromFormat("(DESCRIPTION=(ADDRESS_LIST=(ADDRESS="
            "(PROTOCOL=TCP)(HOST=%s)(PORT=%d)))(CONNECT_DATA=(SID=%s)))",
            host, port, sid);
}


//-----------------------------------------------------------------------------
// Time()
//   Returns a time value suitable for binding.
//-----------------------------------------------------------------------------
static PyObject* Time(
    PyObject* self,			// passthrough argument
    PyObject* args)			// arguments to function
{
    PyErr_SetString(g_NotSupportedErrorException,
            "Oracle does not support time only variables");
    return NULL;
}


//-----------------------------------------------------------------------------
// TimeFromTicks()
//   Returns a time value suitable for binding.
//-----------------------------------------------------------------------------
static PyObject* TimeFromTicks(
    PyObject* self,			// passthrough argument
    PyObject* args)			// arguments to function
{
    PyErr_SetString(g_NotSupportedErrorException,
            "Oracle does not support time only variables");
    return NULL;
}


#ifndef NATIVE_DATETIME
//-----------------------------------------------------------------------------
// Date()
//   Returns a date value suitable for binding.
//-----------------------------------------------------------------------------
static PyObject* Date(
    PyObject* self,			// passthrough argument
    PyObject* args)			// arguments to function
{
    return ExternalDateTimeVar_New(&g_ExternalDateTimeVarType, args, NULL);
}
#endif


//-----------------------------------------------------------------------------
// DateFromTicks()
//   Returns a date value suitable for binding.
//-----------------------------------------------------------------------------
static PyObject* DateFromTicks(
    PyObject* self,			// passthrough argument
    PyObject* args)			// arguments to function
{
#ifdef NATIVE_DATETIME
    return PyDate_FromTimestamp(args);
#else
    struct tm *time;
    time_t ticks;
    int ticksArg;

    if (!PyArg_ParseTuple(args, "i", &ticksArg))
        return NULL;

    ticks = ticksArg;
    time = localtime(&ticks);
    if (time == NULL)
        return PyErr_SetFromErrno(g_DataErrorException);
    return ExternalDateTimeVar_NewFromC(&g_ExternalDateTimeVarType,
            time->tm_year + 1900, time->tm_mon + 1, time->tm_mday, 0, 0, 0, 0);
#endif
}


//-----------------------------------------------------------------------------
// TimestampFromTicks()
//   Returns a date value suitable for binding.
//-----------------------------------------------------------------------------
static PyObject* TimestampFromTicks(
    PyObject* self,			// passthrough argument
    PyObject* args)			// arguments to function
{
#ifdef NATIVE_DATETIME
    return PyDateTime_FromTimestamp(args);
#else
    struct tm *time;
    time_t ticks;
    int ticksArg;

    if (!PyArg_ParseTuple(args, "i", &ticksArg))
        return NULL;

    ticks = ticksArg;
    time = localtime(&ticks);
    if (time == NULL)
        return PyErr_SetFromErrno(g_DataErrorException);
    return ExternalDateTimeVar_NewFromC(&g_ExternalDateTimeVarType,
            time->tm_year + 1900, time->tm_mon + 1, time->tm_mday,
            time->tm_hour, time->tm_min, time->tm_sec, 0);
#endif
}


//-----------------------------------------------------------------------------
//   Declaration of methods supported by this module
//-----------------------------------------------------------------------------
static PyMethodDef g_ModuleMethods[] = {
    { "makedsn", MakeDSN, METH_VARARGS },
#ifndef NATIVE_DATETIME
    { "Date", Date, METH_VARARGS },
#endif
    { "Time", Time, METH_VARARGS },
    { "DateFromTicks", DateFromTicks, METH_VARARGS },
    { "TimeFromTicks", TimeFromTicks, METH_VARARGS },
    { "TimestampFromTicks", TimestampFromTicks, METH_VARARGS },
    { NULL }
};

//-----------------------------------------------------------------------------
// initcx_Oracle()
//   Initialization routine for the shared libary.
//-----------------------------------------------------------------------------
void initcx_Oracle(void)
{
    PyThreadState *threadState;
    PyObject *module;

    if (PyType_Ready(&g_ConnectionType) < 0)
        return;
    if (PyType_Ready(&g_CursorType) < 0)
        return;
    if (PyType_Ready(&g_ErrorType) < 0)
        return;
#ifndef NATIVE_DATETIME
    if (PyType_Ready(&g_ExternalDateTimeVarType) < 0)
        return;
#endif
#ifdef ORACLE_9I
    if (PyType_Ready(&g_SessionPoolType) < 0)
        return;
#endif

    // set the base type for each of the types; this is necessary as Windows
    // does not allow for initialization of structures using anything other
    // than constants
    g_EnvironmentType.ob_type = &PyType_Type;
    g_StringVarType.ob_type = &PyType_Type;
    g_FixedCharVarType.ob_type = &PyType_Type;
    g_RowidVarType.ob_type = &PyType_Type;
    g_BinaryVarType.ob_type = &PyType_Type;
    g_LongStringVarType.ob_type = &PyType_Type;
    g_LongBinaryVarType.ob_type = &PyType_Type;
    g_NumberVarType.ob_type = &PyType_Type;
    g_ExternalLobVarType.ob_type = &PyType_Type;
    g_DateTimeVarType.ob_type = &PyType_Type;
#ifdef ORACLE_9I
    g_TimestampVarType.ob_type = &PyType_Type;
#endif
    g_CLOBVarType.ob_type = &PyType_Type;
    g_BLOBVarType.ob_type = &PyType_Type;
    g_CursorVarType.ob_type = &PyType_Type;

    // initialize module and retrieve the dictionary
    module = Py_InitModule("cx_Oracle", g_ModuleMethods);
    if (!module)
        return;
    g_ModuleDictionary = PyModule_GetDict(module);
    if (!g_ModuleDictionary)
        return;

    // initialize the interpreter state for callbacks
#ifdef WITH_THREAD
    PyEval_InitThreads();
    threadState = PyThreadState_Swap(NULL);
    if (!threadState)
        return;
    g_InterpreterState = threadState->interp;
    PyThreadState_Swap(threadState);
#endif

    // create exception object and add it to the dictionary
    if (SetException(g_ModuleDictionary, &g_WarningException,
            "Warning", PyExc_StandardError) < 0)
        return;
    if (SetException(g_ModuleDictionary, &g_ErrorException,
            "Error", PyExc_StandardError) < 0)
        return;
    if (SetException(g_ModuleDictionary, &g_InterfaceErrorException,
            "InterfaceError", g_ErrorException) < 0)
        return;
    if (SetException(g_ModuleDictionary, &g_DatabaseErrorException,
            "DatabaseError", g_ErrorException) < 0)
        return;
    if (SetException(g_ModuleDictionary, &g_DataErrorException,
            "DataError", g_DatabaseErrorException) < 0)
        return;
    if (SetException(g_ModuleDictionary, &g_OperationalErrorException,
            "OperationalError", g_DatabaseErrorException) < 0)
        return;
    if (SetException(g_ModuleDictionary, &g_IntegrityErrorException,
            "IntegrityError", g_DatabaseErrorException) < 0)
        return;
    if (SetException(g_ModuleDictionary, &g_InternalErrorException,
            "InternalError", g_DatabaseErrorException) < 0)
        return;
    if (SetException(g_ModuleDictionary, &g_ProgrammingErrorException,
            "ProgrammingError", g_DatabaseErrorException) < 0)
        return;
    if (SetException(g_ModuleDictionary, &g_NotSupportedErrorException,
            "NotSupportedError", g_DatabaseErrorException) < 0)
        return;

    // manage the use of the datetime module; in Python 2.4 and up, the native
    // datetime type is used since a C API is exposed; in all other cases, the
    // internal datetime type is used and if the datetime module exists,
    // binding to that type is permitted
#ifdef NATIVE_DATETIME
    PyDateTime_IMPORT;
    if (PyErr_Occurred())
        return;
#else
    module = PyImport_ImportModule("datetime");
    if (module)
        g_DateTimeType = (PyTypeObject*) PyObject_GetAttrString(module,
                "datetime");
    PyErr_Clear();
#endif

    // set up the types that are available
    if (PyDict_SetItemString(g_ModuleDictionary, "Connection",
            (PyObject*) &g_ConnectionType) < 0)
        return;
    if (PyDict_SetItemString(g_ModuleDictionary, "Cursor",
            (PyObject*) &g_CursorType) < 0)
        return;
#ifdef NATIVE_DATETIME
    if (PyDict_SetItemString(g_ModuleDictionary, "Timestamp",
            (PyObject*) PyDateTimeAPI->DateTimeType) < 0)
        return;
    if (PyDict_SetItemString(g_ModuleDictionary, "Date",
            (PyObject*) PyDateTimeAPI->DateType) < 0)
        return;
#else
    if (PyDict_SetItemString(g_ModuleDictionary, "Timestamp",
            (PyObject*) &g_ExternalDateTimeVarType) < 0)
        return;
#endif
#ifdef ORACLE_9I
    if (PyDict_SetItemString(g_ModuleDictionary, "SessionPool",
            (PyObject*) &g_SessionPoolType) < 0)
        return;
#endif
    if (PyDict_SetItemString(g_ModuleDictionary, "_Error",
            (PyObject*) &g_ErrorType) < 0)
        return;

    // the name "connect" is required by the DB API
    if (PyDict_SetItemString(g_ModuleDictionary, "connect",
            (PyObject*) &g_ConnectionType) < 0)
        return;

    // create the basic data types for setting input sizes
    if (PyDict_SetItemString(g_ModuleDictionary, "BINARY",
            (PyObject*) &g_BinaryVarType) < 0)
        return;
    if (PyDict_SetItemString(g_ModuleDictionary, "BLOB",
            (PyObject*) &g_BLOBVarType) < 0)
        return;
    if (PyDict_SetItemString(g_ModuleDictionary, "CLOB",
            (PyObject*) &g_CLOBVarType) < 0)
        return;
    if (PyDict_SetItemString(g_ModuleDictionary, "CURSOR",
            (PyObject*) &g_CursorVarType) < 0)
        return;
#ifdef NATIVE_DATETIME
    if (PyDict_SetItemString(g_ModuleDictionary, "DATETIME",
            (PyObject*) PyDateTimeAPI->DateTimeType) < 0)
        return;
#else
    if (PyDict_SetItemString(g_ModuleDictionary, "DATETIME",
            (PyObject*) &g_ExternalDateTimeVarType) < 0)
        return;
#endif
    if (PyDict_SetItemString(g_ModuleDictionary, "FIXED_CHAR",
            (PyObject*) &g_FixedCharVarType) < 0)
        return;
    if (PyDict_SetItemString(g_ModuleDictionary, "LOB",
            (PyObject*) &g_ExternalLobVarType) < 0)
        return;
    if (PyDict_SetItemString(g_ModuleDictionary, "LONG_BINARY",
            (PyObject*) &g_LongBinaryVarType) < 0)
        return;
    if (PyDict_SetItemString(g_ModuleDictionary, "LONG_STRING",
            (PyObject*) &g_LongStringVarType) < 0)
        return;
    if (PyDict_SetItemString(g_ModuleDictionary, "NUMBER",
            (PyObject*) &g_NumberVarType) < 0)
        return;
    if (PyDict_SetItemString(g_ModuleDictionary, "ROWID",
            (PyObject*) &g_RowidVarType) < 0)
        return;
    if (PyDict_SetItemString(g_ModuleDictionary, "STRING",
            (PyObject*) &g_StringVarType) < 0)
        return;
#ifdef ORACLE_9I
    if (PyDict_SetItemString(g_ModuleDictionary, "TIMESTAMP",
            (PyObject*) &g_TimestampVarType) < 0)
        return;
#endif

    // create constants required by Python DB API 2.0
    if (PyDict_SetItemString(g_ModuleDictionary, "apilevel",
            PyString_FromString("2.0")) < 0)
        return;
    if (PyDict_SetItemString(g_ModuleDictionary, "threadsafety",
            PyInt_FromLong(2)) < 0)
        return;
    if (PyDict_SetItemString(g_ModuleDictionary, "paramstyle",
            PyString_FromString("named")) < 0)
        return;

    // add version and build time for easier support
    if (PyDict_SetItemString(g_ModuleDictionary, "version",
            PyString_FromString("4.1.2")) < 0)
        return;
    if (PyDict_SetItemString(g_ModuleDictionary, "buildtime",
            PyString_FromString(BUILD_TIME)) < 0)
        return;

    // add constants for registering callbacks
    SET_OCI_CONSTANT(SYSDBA)
    SET_OCI_CONSTANT(SYSOPER)
    SET_OCI_CONSTANT(FNCODE_BINDBYNAME)
    SET_OCI_CONSTANT(FNCODE_BINDBYPOS)
    SET_OCI_CONSTANT(FNCODE_DEFINEBYPOS)
    SET_OCI_CONSTANT(FNCODE_STMTEXECUTE)
    SET_OCI_CONSTANT(FNCODE_STMTFETCH)
    SET_OCI_CONSTANT(FNCODE_STMTPREPARE)
    SET_OCI_CONSTANT(UCBTYPE_ENTRY)
    SET_OCI_CONSTANT(UCBTYPE_EXIT)
    SET_OCI_CONSTANT(UCBTYPE_REPLACE)
}

